/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <string.h>

#include <doca_argp.h>
#include <doca_error.h>
#include <doca_dev.h>
#include <doca_sha.h>
#include <doca_log.h>

#include <utils.h>

DOCA_LOG_REGISTER(SHA_CREATE::MAIN);

#define MAX_PCI_ADDRESS_LEN 32	/* max PCI string length */
#define MAX_DATA_LEN 1024	/* max user data length */
#define MIN_DATA_LEN 1		/* min user data length */

/* Configuration struct */
struct sha_cfg {
	char data[MAX_DATA_LEN];		/* data for SHA job */
	char pci_address[MAX_PCI_ADDRESS_LEN];	/* device PCI address */
};

/* Sample's Logic */
doca_error_t sha_create(struct doca_pci_bdf *pci_dev, char *src_buffer);

/*
 * ARGP Callback - Handle PCI device address parameter
 *
 * @param [in]: Input parameter
 * @config [in/out]: Program configuration context
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
pci_address_callback(void *param, void *config)
{
	struct sha_cfg *sha_cfg = (struct sha_cfg *)config;
	char *pci_address = (char *)param;
	int len;

	len = strnlen(pci_address, MAX_PCI_ADDRESS_LEN);
	if (len == MAX_PCI_ADDRESS_LEN) {
		DOCA_LOG_ERR("Pci address too long, max %d", MAX_PCI_ADDRESS_LEN - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strcpy(sha_cfg->pci_address, pci_address);
	return DOCA_SUCCESS;
}

/*
 * ARGP Callback - Handle user data parameter
 *
 * @param [in]: Input parameter
 * @config [in/out]: Program configuration context
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
data_callback(void *param, void *config)
{
	struct sha_cfg *sha_cfg = (struct sha_cfg *)config;
	char *data = (char *)param;
	int len;

	len = strnlen(data, MAX_DATA_LEN);
	if (len == MAX_DATA_LEN || len < MIN_DATA_LEN) {
		DOCA_LOG_ERR("Invalid data length, should be between %d and %d", MIN_DATA_LEN, MAX_DATA_LEN - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strcpy(sha_cfg->data, data);
	return DOCA_SUCCESS;
}

/*
 * Register the command line parameters for the sample.
 *
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
register_sha_params()
{
	doca_error_t result;
	struct doca_argp_param *pci_param, *data_param;

	result = doca_argp_param_create(&pci_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create ARGP param: %s", doca_get_error_string(result));
		return result;
	}
	doca_argp_param_set_short_name(pci_param, "p");
	doca_argp_param_set_long_name(pci_param, "pci-addr");
	doca_argp_param_set_description(pci_param, "PCI device address");
	doca_argp_param_set_callback(pci_param, pci_address_callback);
	doca_argp_param_set_type(pci_param, DOCA_ARGP_TYPE_STRING);
	result = doca_argp_register_param(pci_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register program param: %s", doca_get_error_string(result));
		return result;
	}

	result = doca_argp_param_create(&data_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create ARGP param: %s", doca_get_error_string(result));
		return result;
	}
	doca_argp_param_set_short_name(data_param, "d");
	doca_argp_param_set_long_name(data_param, "data");
	doca_argp_param_set_description(data_param, "user data");
	doca_argp_param_set_callback(data_param, data_callback);
	doca_argp_param_set_type(data_param, DOCA_ARGP_TYPE_STRING);
	result = doca_argp_register_param(data_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register program param: %s", doca_get_error_string(result));
		return result;
	}
	return DOCA_SUCCESS;
}

/*
 * Sample main function
 *
 * @argc [in]: command line arguments size
 * @argv [in]: array of command line arguments
 * @return: EXIT_SUCCESS on success and EXIT_FAILURE otherwise
 */
int
main(int argc, char **argv)
{
	doca_error_t result;
	int exit_status = EXIT_SUCCESS;
	struct sha_cfg sha_cfg = {0};
	struct doca_pci_bdf pcie_dev;

	strcpy(sha_cfg.pci_address, "03:00.0");
	strcpy(sha_cfg.data, "1234567890abcdef");

	result = doca_argp_init("sha_create", &sha_cfg);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to init ARGP resources: %s", doca_get_error_string(result));
		return EXIT_FAILURE;
	}

	result = register_sha_params();
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register ARGP params: %s", doca_get_error_string(result));
		doca_argp_destroy();
		return EXIT_FAILURE;
	}

	result = doca_argp_start(argc, argv);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to parse sample input: %s", doca_get_error_string(result));
		doca_argp_destroy();
		return EXIT_FAILURE;
	}

	result = parse_pci_addr(sha_cfg.pci_address, &pcie_dev);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to parse PCI address: %s", doca_get_error_string(result));
		doca_argp_destroy();
		return EXIT_FAILURE;
	}

	result = sha_create(&pcie_dev, sha_cfg.data);
	if (result != DOCA_SUCCESS)
		exit_status = EXIT_FAILURE;

	doca_argp_destroy();
	return exit_status;
}
